# Copyright (C) 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Abstractions for http request and response messages"""

from StringIO import StringIO
from inteproxy.decompressstream import DecompressStream
from zlib import decompressobj, MAX_WBITS

class HTTPMessage(object):

    """Base class for HTTP Messages in InteProxy.

    Parameters
      headers -- The headers of the message as a mimetools.Message
                 instance.

      infile -- File-object like object from which the body of the
                message can be read.

    The headers object can be accessed as the headers attribute.  The
    body should be accessed through the body property or the message's
    read method.  If the body is read via the body property, the body is
    read into memory from the infile once.  The body can also be set by
    assignment ot the body attribute or the set_body method.  The latter
    can also set the content type.

    If the body is read or set, the headers are adjusted a little.  The
    Content-length is set to the length of the body and any
    Transfer-encoding headers are removed.
    """

    def __init__(self, headers, infile):
        self.headers = headers
        self.infile = infile
        self._body = None
        self._body_stream = None

    def debug_log_message(self, log_function):
        for header, value in self.headers.items():
            note = ""
            if header.lower() in ("authorization", "proxy-authorization"):
                # The values of authorization headers may contain
                # passwords that are only base64 encoded but not
                # encrypted.  To avoid leaking sensitive information to
                # logfiles, remove such information in the debug output
                value = value.split()[0]
                note = " [value shortened in debug output]"
            log_function("header: %s:%r%s", header, value, note)
        if self.body_has_been_read():
            if self._body:
                log_function("body: %r", self._body)
            else:
                log_function("no body")
        else:
            log_function("body may be present but has not been read yet")

    def body_has_been_read(self):
        return self._body != None

    def set_body(self, body, content_type=None):
        self.headers["Content-length"] = str(len(body))
        del self.headers["Transfer-encoding"]
        if content_type is not None:
            self.headers["Content-type"] = content_type
        self._body = body
        self._body_stream = StringIO(self.body)

    def get_body(self):
        self.read_entire_message()
        return self._body

    body = property(get_body, set_body)

    def read_entire_message(self):
        raise NotImplementedError

    def read(self, amount):
        if self._body_stream is not None:
            data = self._body_stream.read(amount)
        else:
            data = self.infile.read(amount)
        return data


class HTTPRequestMessage(HTTPMessage):

    """Represents a HTTP Message for a request.

    Additional parameters:
       method -- The HTTP method of the request (usually 'GET' or 'POST'()

       uri -- The uri of the request

       version -- The http protocol version given in the request

    All of these parameters are available as attributes of the same
    name.
    """

    def __init__(self, method, uri, version, headers, infile):
        super(HTTPRequestMessage, self).__init__(headers, infile)
        self.method = method
        self.uri = uri
        self.version = version

    def debug_log_message(self, log_function):
        log_function("HTTPRequestMessage: %s %s %s",
                     self.method, self.uri, self.version)
        super(HTTPRequestMessage, self).debug_log_message(log_function)

    def read_entire_message(self):
        if self.body_has_been_read():
            return
        body = None
        length = int(self.headers.get("Content-Length", "0"))
        if length:
            # FIXME: check whether length bytes were read
            body = self.infile.read(length)

        if body is not None:
            self.set_body(body)


class HTTPResponseMessage(HTTPMessage):

    """Represents a HTTP Message for a response.

    Additional parameters:
       version -- The http protocol version given in the response

       status -- The HTTP status of the response

       reason -- The reason given in the response

    All of these parameters are available as attributes of the same
    name.
    """

    def __init__(self, version, status, reason, headers, infile):
        super(HTTPResponseMessage, self).__init__(headers, infile)
        self.version = version
        self.status = status
        self.reason = reason
        self.__decompress = False
        self.started_reading = False

    def debug_log_message(self, log_function):
        log_function("HTTPResponseMessage: %s %s %s",
                     self.version, self.status, self.reason)
        super(HTTPResponseMessage, self).debug_log_message(log_function)

    def read_entire_message(self):
        """ Read the entire message and set the messages body.

        If the optional decompressor parameter is given the
        body will be decompressed.
        """
        if not self.body_has_been_read():
            self.set_body(self.read())

    def do_decompress(self):
        """
        Decompress the input stream on read if possible.

        If the content-encoding of the message is either gzip
        or deflate the Content-Encoding and Content-Length
        headers will also be removed.
        """
        if self.__decompress:
            return

        if not self.headers.get("Content-Encoding"):
            return

        if self.headers.get("Content-Encoding") == "deflate":
            self.__decompress = "deflate"
        elif self.headers.get("Content-Encoding") == "gzip":
            self.__decompress = "gzip"

        if self.__decompress:
            # Can decompress the input
            if self.started_reading:
                self.__decompress = False
                raise Exception("do_decompress called after first read")

            del self.headers["Content-Encoding"]
            if self.headers.get("Content-Length"):
                del self.headers["Content-Length"]

    def read(self, amount = -1):
        """
        Read the message up to amount bytes and return a data string of
        length amount.
        If do_decompress was called before this will return the data in
        a decompressed form if possible.
        """

        if self.started_reading == False and amount != 0:
            self.started_reading = True

        if self._body_stream is None:
            if self.__decompress:
                # Create the decompression object
                #
                # On defate decompression -zlib.MAX_WBITS is given to ensure
                # that non RFC confirming responses as they are sent by most
                # http servers are decompressed correctly by ignoring a
                # possibly invlaid header.

                # To decompress gzip with zlib 16 needs to be added to the
                # wbits parameter

                # See the documention of inflateInit2 at
                # http://zlib.net/manual.html

                if self.__decompress == "deflate":
                    self._body_stream = DecompressStream(self.infile,
                            decompressobj(-MAX_WBITS))
                elif self.__decompress == "gzip":
                    self._body_stream = DecompressStream(self.infile,
                            decompressobj(16 + MAX_WBITS))
                else:
                    raise Exception("Invalid decompress method"
                                    " in HTTPResponse")
            else:
                self._body_stream = self.infile

        # Passing the amount value to the read method of the body stream
        # is tricky for the case of an omitted amount value, i.e. one
        # indicating that all the data should be read. The only thing
        # the various implementations seem to have in common is that an
        # omitted amount parameter indicates this. In all cases I have
        # looked at it is possible to pass a value for this parameter
        # with the same meaning, but the value used differs. Some
        # examples:
        #
        # File-like type          Value for omitted amount parameter
        #   file                       negative integer
        #   httplib.HTTPResponse       None
        #   zlib.decompressobj         0
        #
        # Since we may be dealing with any of these and perhaps more,
        # the only way to pass the amount parameter while retaining the
        # correct semantics is to use a possibly empty parameter tuple.
        return self._body_stream.read(*((amount,) if amount > 0 else ()))
