# Copyright (C) 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Tests for InteProxy's SSL validation"""

import unittest

from inteproxy.httpconnection import is_remote_domain_valid


class TestHostNameChecks(unittest.TestCase):

    def match(self, hostname, **cert_dict):
        self.failUnless(is_remote_domain_valid(cert_dict, hostname))

    def nomatch(self, hostname, **cert_dict):
        self.failIf(is_remote_domain_valid(cert_dict, hostname))

    #
    # Test for hostnames that should match
    #

    def test_cn_exact(self):
        self.match("intevation.de",
                   subject=((("commonName", "intevation.de"),),))

    def test_cn_any_single_subdomain(self):
        self.match("www.intevation.de",
                   subject=((("commonName", "*.intevation.de"),),))

    def test_cn_partial_single_subdomain(self):
        self.match("ssl.intevation.de",
                   subject=((("commonName", "s*.intevation.de"),),))

    def test_sn_exact(self):
        self.match("intevation.de",
                   subjectAltName=(("DNS", "*.intevation.de"),
                                   ("DNS", "intevation.de")))

    def test_sn_any_single_subdomain(self):
        self.match("inteproxy-demo.intevation.de",
                   subjectAltName=(("DNS", "*.intevation.de"),
                                   ("DNS", "intevation.de")))

    def test_hostname_lower_case(self):
        self.match("intevation.de",
                   subject=((("commonName", "INtevATioN.DE"),),))

    def test_remote_domain_lower_case(self):
        self.match("inTEvaTIon.de",
                   subject=((("commonName", "INTEVATION.DE"),),))

    def test_both_mixed_case(self):
        self.match("inTEvaTIon.de",
                   subject=((("commonName", "INtevATioN.DE"),),))

    def test_hostname_contains_alphanum(self):
        self.match("intevation1999.de",
                   subject=((("commonName", "intevation1999.de"),),))

    def test_hostname_contains_hypher(self):
        self.match("intevation-1999.de",
                   subject=((("commonName", "intevation-1999.de"),),))

    #
    # Test for hostnames that should NOT match
    #

    def test_cn_too_many_subdomains(self):
        self.nomatch("foo.www.intevation.de",
                     subject=((("commonName", "*.intevation.de"),),))

    def test_cn_matches_only_subdomain_of_hostname(self):
        self.nomatch("www.intevation.de.evil.example.com",
                     subject=((("commonName", "*.intevation.de"),),))


    def test_sn_cn_but_only_cn_matches(self):
        self.nomatch("www.intevation.de.evil.example.com",
                     subject=((("commonName", "www.intevation.de"),),),
                     subjectAltName=(("DNS", "ssl.intevation.de"),))

    def test_hostname_start_with_wrong_char(self):
        self.nomatch("-intevation.de",
                     subjectAltName=(("DNS", "-intevation.de"),))

    def test_hostname_ends_with_wrong_char(self):
        self.nomatch("intevation-.de",
                     subjectAltName=(("DNS", "intevation-.de"),))

    def test_hostname_ends_with_wrong_newline(self):
        self.nomatch("intevation\n.de",
                     subjectAltName=(("DNS", "intevation\n.de"),))

    def test_hostname_contains_invalid_char(self):
        self.nomatch("de-intevation.de",
                     subjectAltName=(("DNS", "dem?-intevation.de"),))

    def test_hostname_contains_empty_component(self):
        self.nomatch("..de",
                     subjectAltName=(("DNS", "..de"),))

